#define __ERROR_C
#include <stdio.h>
#include <stdarg.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>

#include "error.h"

char errormsg[ERROR_MSG_BUFLEN] = "";

const char error_malloc[] = "Konnte keinen Speicher reservieren.";
pthread_mutex_t mutex_error = PTHREAD_MUTEX_INITIALIZER;


static inline pid_t _gettid(void)
	{
	return (pid_t)syscall(SYS_gettid);
	}


void error_lock(void)
	{
	pthread_mutex_lock(&mutex_error);
	}


void error_unlock(void)
	{
	pthread_mutex_unlock(&mutex_error);
	}


void error_append(const char *file, unsigned int line, const char *function, const char *format, ...)
	{
	va_list args;
	va_start (args, format);

	pthread_mutex_lock(&mutex_error);

	if(_gettid() == getpid())
		{
		fprintf(	stdout,
					"%s(%u) in Function %s: ",
					file,
					line,
					function);
		}
	else
		{
		fprintf(	stdout,
					"Thread %d - %s(%u) in Function %s: ",
					(int)_gettid(),
					file,
					line,
					function);
		}

	vfprintf(	stdout,
				format, args);
	fprintf(	stdout,
				"\n");

	pthread_mutex_unlock(&mutex_error);

	va_end(args);
	}
/*
void error_append(const char *file, unsigned int line, const char *function, const char *format, ...)
	{
	va_list args;
	va_start (args, format);

	pthread_mutex_lock(&mutex_error);

	if(_gettid() == getpid())
		{
		snprintf(	&errormsg[strlen(errormsg)],
					ERROR_MSG_BUFLEN -strlen(errormsg),
					"%s(%u) in Function %s: ",
					file,
					line,
					function);
		}
	else
		{
		snprintf(	&errormsg[strlen(errormsg)],
					ERROR_MSG_BUFLEN -strlen(errormsg),
					"Thread %d - %s(%u) in Function %s: ",
					(int)_gettid(),
					file,
					line,
					function);
		}

	vsnprintf(	&errormsg[strlen(errormsg)],
				ERROR_MSG_BUFLEN -strlen(errormsg),
				format, args);
	snprintf(	&errormsg[strlen(errormsg)],
				ERROR_MSG_BUFLEN -strlen(errormsg),
				"\n");

	pthread_mutex_unlock(&mutex_error);

	va_end(args);
	}
*/

void error_delete(void)
	{
	pthread_mutex_lock(&mutex_error);
	errormsg[0] = 0;
	pthread_mutex_unlock(&mutex_error);
	}


int error_is_pending(void)
	{
	int ret;

	pthread_mutex_lock(&mutex_error);

	if(strlen(errormsg))
		ret = 1;
	else
		ret = 0;

	pthread_mutex_unlock(&mutex_error);

	return ret;
	}


void error_print(FILE *out)
	{
	pthread_mutex_lock(&mutex_error);

	if(errormsg[0] != 0)
		{
		fprintf(out, "%s", errormsg);
		errormsg[0] = 0;
		}

	pthread_mutex_unlock(&mutex_error);
	}
