#include <sys/socket.h>
#include <arpa/inet.h>
#include <linux/if_ether.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <stdint.h>
#include <inttypes.h>
#include <linux/if_packet.h>
#include <poll.h>

#include "error.h"
#include "helper.h"
#include "eth_hw.h"


// ---------- Speicherverwaltung ----------


#define MEM_ENTRY_COUNT	10
#define MEM_ENTRY_SIZE	1600

int eth_hw_memsize = MEM_ENTRY_SIZE;

static uint8_t mem_isallocated[MEM_ENTRY_COUNT];
static uint8_t mem_pool[MEM_ENTRY_COUNT *MEM_ENTRY_SIZE];


void *eth_hw_alloc(void)
	{
	for(int i=0;i<NELEMS(mem_isallocated);i++)
		{
		if(mem_isallocated[i] == 0)
			{
			mem_isallocated[i] = 1;
			return &(mem_pool[i*MEM_ENTRY_SIZE]);
			}
		}
	ERROR_APPEND("Speicher komplett belegt.\n");
	exit(1);
	return NULL;
	}


void eth_hw_free(void *data)
	{
	uintptr_t offset = (uintptr_t)data - (uintptr_t)mem_pool;
	uintptr_t index = offset / MEM_ENTRY_SIZE;
	if(	(uintptr_t)data < (uintptr_t)mem_pool ||
		(uintptr_t)data >= (uintptr_t)mem_pool + sizeof(mem_pool) ||
		offset % MEM_ENTRY_SIZE != 0 || // Adresse ist nicht der Anfang eines Buffers
		mem_isallocated[index] == 0) // Noch nicht vergeben
		{
		// Ungültiger free
		ERROR_APPEND("Ungültiger free\n");
		exit(1);
		//for(;;)asm volatile("NOP");
		}
	mem_isallocated[index] = 0;
	}


// Offset der Adresse vom Block wird toleriert
void eth_hw_free_offset(void *data)
	{
	uintptr_t offset = (uintptr_t)data - (uintptr_t)mem_pool;
	uintptr_t index = offset / MEM_ENTRY_SIZE;
	if(	(uintptr_t)data < (uintptr_t)mem_pool ||
		(uintptr_t)data >= (uintptr_t)mem_pool + sizeof(mem_pool) ||
		mem_isallocated[index] == 0) // Noch nicht vergeben
		{
		// Ungültiger free
		ERROR_APPEND("Ungültiger free\n");
		exit(1);
		//for(;;)asm volatile("NOP");
		}
	mem_isallocated[index] = 0;
	}


// ---------- Socket ----------

static int sockfd = -1;
static const char devicename[] = "ethdummy";


void eth_hw_init(void)
	{
	if ((sockfd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_IP))) != -1)
		{
		struct ifreq ifr;
		memset(&ifr, 0x00, sizeof(ifr));
		strncpy(ifr.ifr_name, devicename, IFNAMSIZ);
		if (ioctl(sockfd, SIOCGIFINDEX, &ifr) == 0)
			{
			//int ethindex = ifr.ifr_ifindex;
			struct sockaddr_ll sll;
			memset(&sll, 0, sizeof(sll));
			sll.sll_family = AF_PACKET;
			sll.sll_ifindex = ifr.ifr_ifindex;
			sll.sll_protocol = htons(ETH_P_ALL);
			if((bind(sockfd, (struct sockaddr *)&sll, sizeof(sll))) == 0)
				{
				struct timeval read_timeout;
				read_timeout.tv_sec = 0;
				read_timeout.tv_usec = 1;
				if(setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, &read_timeout, sizeof(read_timeout)) == 0)
					{
					// Nur an das Interface binden
					//if(setsockopt(sockfd, SOL_SOCKET, SO_BINDTODEVICE, devicename, strlen(devicename) + 1) == 0)
						{
						// promiscuos mode
						// SIOCGIFFLAGS	0x8913 get flags
						// SIOCSIFFLAGS	0x8914 set flags
						struct ifreq ethreq;
						strncpy(ethreq.ifr_name, devicename, IF_NAMESIZE);
						if(ioctl(sockfd, SIOCGIFFLAGS, &ethreq) == 0)
							{
							ethreq.ifr_flags |= IFF_PROMISC;
							if(ioctl(sockfd, SIOCSIFFLAGS, &ethreq) == 0)
								{
								return;
								}
							else
								{
								ERROR_APPEND_ERRNO("\"ioctl(SIOCSIFFLAGS)\" meldet Fehler.");
								}
							}
						else
							{
							ERROR_APPEND_ERRNO("\"ioctl(SIOCGIFFLAGS)\" meldet Fehler.");
							}
						}
//					else
						{
//						ERROR_APPEND_ERRNO("\"setsockopt(SO_BINDTODEVICE)\" meldet Fehler.");
						}
					}
				else
					{
					ERROR_APPEND_ERRNO("\"setsockopt(SO_RCVTIMEO)\" meldet Fehler.");
					}
				}
			else
				{
				ERROR_APPEND_ERRNO("\"bind\" meldet Fehler.");
				}
			}
		else
			{
			ERROR_APPEND_ERRNO("\"ioctl(SIOCGIFINDEX)\" meldet Fehler.");
			}
		close(sockfd);
		sockfd = -1;
		}
	else
		{
		ERROR_APPEND_ERRNO("\"socket\" meldet Fehler.");
		}
	exit(1);
	}


void eth_hw_uninit(void)
	{
	close(sockfd);
	sockfd = -1;
	}


int eth_hw_read(void **buffer)
	{
	//~ struct sockaddr_ll saddr;
	//~ socklen_t saddrlen = sizeof(saddr);
	//~ memset((void*)&saddr, 0, sizeof(saddr));
	//~ saddr.sll_family = PF_PACKET;   
	//~ saddr.sll_ifindex = ethindex;
	//~ saddr.sll_halen = ETH_ALEN;
	//~ ret = recvfrom(sockfd, buffer, sizeof(buffer), 0, (void*)&saddr, &saddrlen);
	*buffer = eth_hw_alloc();
	if(buffer == NULL)
		{
		ERROR_APPEND("Fehler beim Reservieren von Speicher.");
		return 0;
		}

	int ret = recvfrom(sockfd, *buffer, eth_hw_memsize, 0, NULL, NULL);
	if(ret == 0)
		{
		ERROR_APPEND("\"recvfrom\" meldet geschlossene Verbindung.");
		exit(1);
		}
	if(ret < 0)
		{
		if(errno == EAGAIN)
			{
			// Kein Paket empfangen
			eth_hw_free(*buffer);
			return 0;
			}
		ERROR_APPEND("\"recvfrom\" meldet Fehler.");
		exit(1);
		}
	return ret;
	}


void eth_hw_send(void *data, int length)
	{
	int ret = sendto(sockfd, data, length, 0, NULL, 9);
	if(ret != length)
		{
		ERROR_APPEND("\"sendto\" meldet Fehler.");
		exit(1);
		}
	eth_hw_free(data);
	}
