#include <sys/socket.h>
#include <arpa/inet.h>
#include <linux/if_ether.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <stdint.h>
#include <inttypes.h>
#include <linux/if_packet.h>
#include <poll.h>
#include <linux/if.h>
#include <linux/if_tun.h>
#include <fcntl.h>


#include "error.h"
#include "helper.h"
#include "eth_hw.h"


// ---------- Speicherverwaltung ----------


#define MEM_ENTRY_COUNT	10
#define MEM_ENTRY_SIZE	1600

int eth_hw_memsize = MEM_ENTRY_SIZE;

static uint8_t mem_isallocated[MEM_ENTRY_COUNT];
static uint8_t mem_pool[MEM_ENTRY_COUNT *MEM_ENTRY_SIZE];


void *eth_hw_alloc(void)
	{
	for(int i=0;i<NELEMS(mem_isallocated);i++)
		{
		if(mem_isallocated[i] == 0)
			{
			mem_isallocated[i] = 1;
			return &(mem_pool[i*MEM_ENTRY_SIZE]);
			}
		}
	ERROR_APPEND("Speicher komplett belegt.\n");
	exit(1);
	return NULL;
	}


void eth_hw_free(void *data)
	{
	uintptr_t offset = (uintptr_t)data - (uintptr_t)mem_pool;
	uintptr_t index = offset / MEM_ENTRY_SIZE;
	if(	(uintptr_t)data < (uintptr_t)mem_pool ||
		(uintptr_t)data >= (uintptr_t)mem_pool + sizeof(mem_pool) ||
		offset % MEM_ENTRY_SIZE != 0 || // Adresse ist nicht der Anfang eines Buffers
		mem_isallocated[index] == 0) // Noch nicht vergeben
		{
		// Ungültiger free
		ERROR_APPEND("Ungültiger free\n");
		*(uint32_t*)0 = 1;
		exit(1);
		//for(;;)asm volatile("NOP");
		}
	mem_isallocated[index] = 0;
	}


// Offset der Adresse vom Block wird toleriert
void eth_hw_free_offset(void *data)
	{
	uintptr_t offset = (uintptr_t)data - (uintptr_t)mem_pool;
	uintptr_t index = offset / MEM_ENTRY_SIZE;
	if(	(uintptr_t)data < (uintptr_t)mem_pool ||
		(uintptr_t)data >= (uintptr_t)mem_pool + sizeof(mem_pool) ||
		mem_isallocated[index] == 0) // Noch nicht vergeben
		{
		// Ungültiger free
		ERROR_APPEND("Ungültiger free\n");
		*(uint32_t*)0 = 1;
		exit(1);
		//for(;;)asm volatile("NOP");
		}
	mem_isallocated[index] = 0;
	}


// ---------- Socket ----------

static int tapfd = -1;
static const char devicename[] = "ethdummy";
static const char blockdevicename[] = "/dev/net/tun";


void eth_hw_uninit(void)
	{
	if(tapfd > 0)
		{
		close(tapfd);
		tapfd = -1;
		}
	}


void eth_hw_init(void)
	{
	if((tapfd = open(blockdevicename, O_RDWR)) >= 0 )
		{
		atexit(eth_hw_uninit);

		struct ifreq ifr;
		memset(&ifr, 0, sizeof(ifr));
// Flags:
// IFF_TUN   - TUN device (no Ethernet headers) 
// IFF_TAP   - TAP device  
// IFF_NO_PI - Do not provide packet information (sonst sind 4 Bytes am Begin jeder Nachricht)
		ifr.ifr_flags = IFF_TAP | IFF_NO_PI; 
		strncpy(ifr.ifr_name, devicename, IFNAMSIZ);
		if(ioctl(tapfd, TUNSETIFF, (void *) &ifr) == 0)
			{
			//strcpy(dev, ifr.ifr_name);
			
			char sysstr[128];
			sprintf(sysstr, "ip addr add 172.16.1.1/24 dev %s", devicename); system(sysstr);
			sprintf(sysstr, "ip link set %s up", devicename); system(sysstr);			
			
			return;
			}
		close(tapfd);
		}
	else
		{
		ERROR_APPEND("Konnte \"%s\" nicht öffnen.", blockdevicename);
		}
	exit(1);
	}


int eth_hw_read(void **buffer)
	{
	*buffer = eth_hw_alloc();
	if(buffer == NULL)
		{
		ERROR_APPEND("Fehler beim Reservieren von Speicher.");
		return 0;
		}

	int ret = read(tapfd, *buffer, eth_hw_memsize);
	if(ret == 0)
		{
		ERROR_APPEND("\"read\" meldet geschlossene Verbindung.");
		exit(1);
		}
	if(ret < 0)
		{
		if(errno == EAGAIN)
			{
			// Kein Paket empfangen
			eth_hw_free(*buffer);
			return 0;
			}
		ERROR_APPEND("\"read\" meldet Fehler.");
		exit(1);
		}
	return ret;
	}


void eth_hw_send(void *data, int length)
	{
	int ret = write(tapfd, data, length);
	if(ret != length)
		{
		ERROR_APPEND("\"sendto\" meldet Fehler.");
		exit(1);
		}
	eth_hw_free(data);
	}
